from .verification import verify_response, verify_signed_in, verify_config_variables, verify_rest_api_version, \
    verify_api_method_exists, verify_connection
from .validation import validate_schedule_state_override



"""
Functions to add:
get_all_subscription_fields
get_sites_dataframe
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages, flatten_dict_column


def get_all_subscription_fields(conn):
    all_subscriptions = extract_pages(conn.query_subscriptions, parameter_dict={'fields': 'fields=_all_'})
    return all_subscriptions


def get_subscriptions_dataframe(conn):
    subscriptions_df = pd.DataFrame(get_all_subscription_fields(conn))
    if not subscriptions_df.empty:
        subscriptions_df = flatten_dict_column(subscriptions_df, keys=['id', 'type'], col_name='content')
        subscriptions_df = flatten_dict_column(subscriptions_df, keys=['id', 'name'], col_name='schedule')
        subscriptions_df = flatten_dict_column(subscriptions_df, keys=['id', 'name'], col_name='user')
    return subscriptions_df



from tableau_api_lib.api_endpoints import BaseEndpoint


class AuthEndpoint(BaseEndpoint):
    def __init__(
        self,
        ts_connection,
        sign_in=False,
        sign_out=False,
        switch_site=False,
        get_server_info=False,
        revoke_admin_pat=False,
        parameter_dict=None,
    ):
        """
        Builds the auth endpoint for Tableau Server REST API auth methods.
        :param class ts_connection: the Tableau Server connection object.
        :param bool sign_in: True if signing in, False otherwise.
        :param bool sign_out: True if signing out, False otherwise.
        :param bool switch_site: True if switching site, False otherwise.
        :param bool get_server_info: True if getting server info, False otherwise.
        :param bool revoke_admin_pat: True if revoking all administrator personal access tokens, False otherwise.
        :param dict parameter_dict: dictionary of URL parameters to append; the value in each key-value pair is the literal
        text that will be appended to the URL endpoint.
        """

        super().__init__(ts_connection)
        self._sign_in = sign_in
        self._sign_out = sign_out
        self._switch_site = switch_site
        self._get_server_info = get_server_info
        self._revoke_admin_pat = revoke_admin_pat
        self._parameter_dict = parameter_dict

    @property
    def base_auth_url(self) -> str:
        return "{0}/api/{1}/auth".format(self._connection.server, self._connection.api_version)

    @property
    def base_sign_in_url(self) -> str:
        return "{0}/signin".format(self.base_auth_url)

    @property
    def base_sign_out_url(self) -> str:
        return "{0}/signout".format(self.base_auth_url)

    @property
    def base_switch_site_url(self) -> str:
        return "{0}/switchSite".format(self.base_auth_url)

    @property
    def base_server_info_url(self) -> str:
        return "{0}/api/{1}/serverinfo".format(self._connection.server, self._connection.api_version)

    @property
    def base_revoke_admin_pat_url(self) -> str:
        return "{0}/serverAdminAccessTokens".format(self.base_auth_url)

    def get_endpoint(self):
        url = None
        if self._sign_in and not (self._sign_out or self._switch_site):
            url = self.base_sign_in_url
        elif self._sign_out and not (self._sign_in or self._switch_site):
            url = self.base_sign_out_url
        elif self._switch_site and not (self._sign_in or self._sign_out):
            url = self.base_switch_site_url
        elif self._get_server_info:
            url = self.base_server_info_url
        elif self._revoke_admin_pat:
            url = self.base_revoke_admin_pat_url

        if url:
            return self._append_url_parameters(url)
        else:
            self._invalid_parameter_exception()



import urllib
from typing import List, Optional

import pandas as pd


def flatten_dict_column(
    df: pd.DataFrame, keys: List[str], col_name: str, add_col_prefix: Optional[bool] = True
) -> pd.DataFrame:
    """
    Extract the specified Pandas DataFrame dict column ('col_name'), creating new columns from the given keys.
    :param pd.DataFrame df: the Pandas DataFrame whose dict column will be flattened
    :param list keys: the keys of the dict column which will be extracted into new columns
    :param str col_name: the name of the Pandas DataFrame column whose content will be flattened
    :param bool add_col_prefix: adds the original 'col_name' value as a prefix to generated columns if True
    :return: pd.DataFrame
    """
    try:
        for key in keys:
            if add_col_prefix is True:
                df[col_name + "_" + key] = df[col_name].apply(lambda col: col[key])
            elif add_col_prefix is False:
                df[key] = df[col_name].apply(lambda col: col[key])
            else:
                raise ValueError("The 'add_col_prefix' value must be set to either True or False.")
        df.drop(columns=[col_name], inplace=True)
        return df
    except KeyError:
        # TODO(elliott): change the `col_name` reference below be the key name when the key is what's missing
        raise KeyError(f"No column named '{col_name}' was found in the DataFrame provided.")


def flatten_dict_list_column(df: pd.DataFrame, col_name: str) -> pd.DataFrame:
    """
    Extract the dicts contained within the specified list column ('col_name'), creating new rows and columns.
    :param pd.DataFrame df: the Pandas DataFrame whose list of dicts (column) will be flattened
    :param str col_name: the name of the Pandas DataFrame column whose content will be flattened
    :return: pd.DataFrame
    """
    flattened_col_df = pd.DataFrame()
    for index, row in df.iterrows():
        temp_df = pd.DataFrame(row[col_name])
        temp_df.index = [index] * temp_df.shape[0]
        flattened_col_df = pd.concat([flattened_col_df, temp_df])
    new_df = df.drop(columns=[col_name]).join(flattened_col_df, how="inner")
    return new_df


def get_server_netloc(server_address: str):
    return urllib.parse.urlparse(server_address).netloc



from tableau_api_lib.api_requests import BaseRequest


class CreateWebhookRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests creating webhooks.
    :param class ts_connection: the Tableau Server connection object
    :param str webhook_name: the name of the webhook being created
    :param str webhook_source_api_event_name: one of the valid documented source api event names
    :param str http_request_method: the http request method [GET, POST, DELETE]
    :param str url: the destination URL for the webhook; must be https and have a valid certificate
    """
    def __init__(self,
                 ts_connection,
                 webhook_name=None,
                 webhook_source_api_event_name=None,
                 http_request_method=None,
                 url=None):

        super().__init__(ts_connection)
        self._webhook_name = webhook_name
        self._webhook_source_api_event_name = webhook_source_api_event_name
        self._http_request_method = str(http_request_method).upper()
        self._url = url
        self._validate_inputs()

    def _validate_inputs(self):
        valid = True
        if self._webhook_source_api_event_name not in self.valid_webhook_source_api_event_names:
            valid = False
        if self._http_request_method not in self.valid_http_request_methods:
            valid = False
        if not valid:
            raise ValueError(f'The webhook source API event name "{self._webhook_source_api_event_name} is invalid."')

    @property
    def valid_webhook_source_api_event_names(self):
        return [
            'webhook-source-event-datasource-refresh-started',
            'webhook-source-event-datasource-refresh-succeeded',
            'webhook-source-event-datasource-refresh-failed',
            'webhook-source-event-datasource-updated',
            'webhook-source-event-datasource-created',
            'webhook-source-event-datasource-deleted',
            'webhook-source-event-workbook-updated',
            'webhook-source-event-workbook-created',
            'webhook-source-event-workbook-deleted',
            'webhook-source-event-workbook-refresh-started',
            'webhook-source-event-workbook-refresh-succeeded',
            'webhook-source-event-workbook-refresh-failed'
        ]

    @property
    def valid_http_request_methods(self):
        return ['GET', 'POST', 'DELETE']

    @property
    def required_webhook_param_keys(self):
        return ['name']

    @property
    def required_webhook_destination_param_keys(self):
        return ['method', 'url']

    @property
    def required_webhook_param_values(self):
        return [self._webhook_name]

    @property
    def required_webhook_destination_param_values(self):
        return [self._http_request_method, self._url]

    def base_create_webhook_request(self):
        if all(self.required_webhook_param_values) and all(self.required_webhook_destination_param_values):
            self._request_body.update({'webhook': {}})
            self._request_body['webhook'].update(
                self._get_parameters_dict(self.required_webhook_param_keys,
                                          self.required_webhook_param_values))
            self._request_body['webhook'].update({'webhook-source': {}})
            self._request_body['webhook']['webhook-source'].update({self._webhook_source_api_event_name: {}})
            self._request_body['webhook'].update({'webhook-destination': {}})
            self._request_body['webhook']['webhook-destination'].update({'webhook-destination-http': {}})
            self._request_body['webhook']['webhook-destination']['webhook-destination-http'].update(
                self._get_parameters_dict(self.required_webhook_destination_param_keys,
                                          self.required_webhook_destination_param_values))
        else:
            self._invalid_parameter_exception()

        return self._request_body

    def get_request(self):
        return self.base_create_webhook_request()



"""This module defines helper functions for querying REST API data for groups."""

from typing import Any, Dict, List, Union

import pandas as pd
from typeguard import typechecked

from tableau_api_lib import TableauServerConnection
from tableau_api_lib.utils import extract_pages
from tableau_api_lib.utils.querying import get_users_dataframe


@typechecked
def get_all_group_fields(conn: TableauServerConnection) -> List[Dict[str, Any]]:
    """Returns details for all groups in the Tableau Server environment, including all queryable fields."""
    all_groups = extract_pages(conn.query_groups, parameter_dict={"fields": "fields=_all_"})
    return all_groups


@typechecked
def get_group_users(conn: TableauServerConnection, group_id: str) -> List[Dict[str, Any]]:
    """Returns details of users belonging to the specified Tableau group."""
    all_group_users = extract_pages(
        conn.get_users_in_group, content_id=group_id, parameter_dict={"fields": "fields=_all_"}
    )
    return all_group_users


@typechecked
def get_all_group_names(conn: TableauServerConnection) -> List[str]:
    """Returns a list of all groups available in the Tableau Server environment."""
    all_groups = get_all_group_fields(conn)
    all_groupnames = [group["name"] for group in all_groups]
    return all_groupnames


@typechecked
def get_all_group_domain_names(conn: TableauServerConnection) -> List[str]:
    """Returns a list of domain names for all groups in the Tableau Server environment."""
    all_groups = get_all_group_fields(conn)
    all_group_roles = [group["domain"]["name"] for group in all_groups]
    return all_group_roles


@typechecked
def get_groups_dataframe(conn: TableauServerConnection) -> pd.DataFrame:
    """Returns a Pandas DataFrame containing details for all groups on the active site."""
    groups_df = pd.DataFrame(get_all_group_fields(conn))
    groups_df["domain"] = groups_df["domain"].apply(lambda x: x["name"])
    return groups_df


@typechecked
def get_group_users_dataframe(conn: TableauServerConnection, group_id: str) -> pd.DataFrame:
    """Returns a Pandas DataFrame describing all users belonging to the specified Tableau group."""
    group_users_df = pd.DataFrame(get_group_users(conn, group_id))
    users_df = get_users_dataframe(conn)
    users_columns = list(users_df.columns)
    if group_users_df.empty is True:
        return group_users_df
    else:
        all_group_users_df = group_users_df.merge(
            users_df, how="left", left_on="id", right_on="id", suffixes=("delete", None)
        )
        return all_group_users_df[users_columns]


@typechecked
def get_groups_for_a_user_dataframe(conn: TableauServerConnection, user_id: str) -> pd.DataFrame:
    """Returns a Pandas DataFrame containing all groups a user belongs to."""
    groups_for_user_df = pd.DataFrame(extract_pages(conn.get_groups_for_a_user, content_id=user_id))
    return groups_for_user_df



from tableau_api_lib.api_endpoints import BaseEndpoint


class WebhookEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 create_webhook=False,
                 query_webhook=False,
                 query_webhooks=False,
                 test_webhook=False,
                 delete_webhook=False,
                 webhook_id=None,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API webhook methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool create_webhook: True if creating a webhook, False otherwise
        :param bool query_webhook: True if querying a specific webhook, False otherwise
        :param bool query_webhooks: True if querying all webhooks, False otherwise
        :param bool test_webhook: True if updating a specific webhook, False otherwise
        :param bool delete_webhook: True if deleting a specific webhook, False otherwise
        :param str webhook_id: the webhook ID
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._create_webhook = create_webhook
        self._query_webhook = query_webhook
        self._query_webhooks = query_webhooks
        self._test_webhook = test_webhook
        self._delete_webhook = delete_webhook
        self._webhook_id = webhook_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._create_webhook,
            self._query_webhook,
            self._query_webhooks,
            self._test_webhook,
            self._delete_webhook
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_webhook_url(self):
        return "{0}/api/{1}/sites/{2}/webhooks".format(self._connection.server,
                                                       self._connection.api_version,
                                                       self._connection.site_id)

    @property
    def base_webhook_id_url(self):
        return "{0}/{1}".format(self.base_webhook_url,
                                self._webhook_id)

    @property
    def base_webhook_test_url(self):
        return "{0}/test".format(self.base_webhook_id_url)

    def get_endpoint(self):
        if self._webhook_id:
            if self._test_webhook:
                url = self.base_webhook_test_url
            else:
                url = self.base_webhook_id_url
        else:
            url = self.base_webhook_url
        if url:
            return self._append_url_parameters(url)
        else:
            self._invalid_parameter_exception()



from typing import List

from tableau_api_lib.api_requests import BaseRequest
from tableau_api_lib.tableau_server_connection import TableauServerConnection


class CreateGroupRequest(BaseRequest):
    """Builds the request body for creating groups using Tableau's REST API."""

    def __init__(self, ts_connection: TableauServerConnection, **kwargs):
        super().__init__(ts_connection)
        self._new_group_name = kwargs.get("new_group_name")
        self._active_directory_group_name = kwargs.get("active_directory_group_name")
        self._active_directory_domain_name = kwargs.get("active_directory_domain_name")
        self._minimum_site_role = kwargs.get("minimum_site_role")
        self._license_mode = kwargs.get("license_mode")
        self._active_directory_source = "ActiveDirectory"
        self._validate_inputs()

    def _validate_inputs(self) -> None:
        valid = True
        if self._license_mode and not self._minimum_site_role:
            valid = False
        if self._active_directory_group_name and not self._active_directory_domain_name:
            valid = False
        if self._active_directory_group_name and not (self._minimum_site_role and self._license_mode):
            raise ValueError("Active Directory groups require setting `minimum_site_role` role and `license_mode`.")
        if valid is False:
            self._invalid_parameter_exception()

    @property
    def is_active_directory_request(self) -> bool:
        return all([self._active_directory_domain_name, self._active_directory_group_name])

    @property
    def valid_site_roles(self) -> List[str]:
        # TODO(elliott): move this type of configuration to enumeration (do for all request classes)
        return [
            "Unlicensed",
            "Viewer",
            "Explorer",
            "ExplorerCanPublish",
            "Creator",
            "SiteAdministratorExplorer",
        ]

    @property
    def valid_license_modes(self) -> List[str]:
        return ["onSync", "onLogin"]

    @property
    def required_local_group_param_keys(self) -> List[str]:
        return ["name"]

    @property
    def optional_local_source_param_keys(self) -> List[str]:
        return ["grantLicenseMode", "minimumSiteRole"]

    @property
    def required_local_group_param_values(self) -> List[str]:
        return [self._new_group_name]

    @property
    def optional_local_group_param_values(self) -> List[str]:
        return [self._license_mode, self._minimum_site_role]

    @property
    def required_active_directory_group_param_keys(self) -> List[str]:
        return ["name", "source", "domainName"]

    @property
    def optional_active_directory_group_param_keys(self) -> List[str]:
        return ["grantLicenseMode", "siteRole"]

    @property
    def required_active_directory_group_param_values(self) -> List[str]:
        return [
            self._new_group_name,
            self._active_directory_source,
            self._active_directory_domain_name,
        ]

    @property
    def optional_active_directory_source_param_values(self) -> List[str]:
        return [self._license_mode, self._minimum_site_role]

    def base_create_local_group_request(self) -> dict:
        if all(self.required_local_group_param_values):
            self._request_body.update({"group": {}})
            self._request_body["group"].update(
                # TODO(elliott): refactor this process for all classes to iterate dict.items()
                self._get_parameters_dict(
                    self.required_local_group_param_keys,
                    self.required_local_group_param_values,
                )
            )
        else:
            self._invalid_parameter_exception()
        return self._request_body

    def modified_create_local_group_request(self) -> dict:
        if any(self.optional_local_group_param_values):
            self._request_body["group"].update(
                self._get_parameters_dict(
                    self.optional_local_source_param_keys,
                    self.optional_local_group_param_values,
                )
            )
        return self._request_body

    def base_create_active_directory_group_request(self) -> dict:
        if all(self.required_active_directory_group_param_values):
            self._request_body.update({"group": {}})
            self._request_body["group"].update(
                self._get_parameters_dict(
                    self.required_active_directory_group_param_keys[:1],
                    self.required_active_directory_group_param_values[:1],
                )
            )
            self._request_body["group"].update({"import": {}})
            self._request_body["group"]["import"].update(
                self._get_parameters_dict(
                    self.required_active_directory_group_param_keys[1:],
                    self.required_active_directory_group_param_values[1:],
                )
            )
        else:
            self._invalid_parameter_exception()
        return self._request_body

    def modified_create_active_directory_request(self) -> dict:
        if any(self.optional_active_directory_source_param_values):
            self._request_body["group"]["import"].update(
                self._get_parameters_dict(
                    self.optional_active_directory_group_param_keys,
                    self.optional_active_directory_source_param_values,
                )
            )
        return self._request_body

    def get_request(self):
        if self.is_active_directory_request:
            self.base_create_active_directory_group_request()
            return self.modified_create_active_directory_request()
        else:
            self.base_create_local_group_request()
            return self.modified_create_local_group_request()



from typing import Any, Dict, List, Union

from tableau_api_lib.api_requests import BaseRequest
from tableau_api_lib.config.api_requests import SiteConfig
from tableau_api_lib.tableau_server_connection import TableauServerConnection


class CreateSiteRequest(BaseRequest):
    """Builds the body for Tableau Server REST API requests creating new Tableau sites."""

    def __init__(
        self, ts_connection: TableauServerConnection, site_name: str, content_url: str, **kwargs: Dict[str, Any]
    ):

        super().__init__(ts_connection)
        self._site_name = site_name
        self._content_url = content_url
        self._kwargs = kwargs
        self._validate_inputs()
        self.base_create_site_request()

    @property
    def valid_extract_encryption_modes(self) -> List[Union[str, None]]:
        return ["enforced", "enabled", "disabled", None]

    def _validate_inputs(self) -> None:
        invalid_inputs = []
        if self._kwargs.get("extract_encryption_mode") not in self.valid_extract_encryption_modes:
            invalid_inputs.append(self._kwargs.get("extract_encryption_mode"))
        if any(invalid_inputs):
            self._invalid_parameter_exception()

    @property
    def optional_param_keys(self) -> List[str]:
        return SiteConfig.OPTIONAL_PARAM_KEYS.value

    @property
    def optional_param_values(self) -> List[Union[str, None]]:
        params = SiteConfig.OPTIONAL_PARAM_VALUES.value
        return [self._kwargs.get(param) for param in params]

    def base_create_site_request(self) -> Dict[str, Any]:
        self._request_body.update(
            {
                "site": {
                    "name": self._site_name,
                    "contentUrl": self._content_url,
                    "adminMode": self._kwargs.get("admin_mode"),
                }
            }
        )
        return self._request_body

    def modified_create_site_request(self) -> Dict[str, Any]:
        if self._kwargs.get("user_quota") and self._kwargs.get("admin_mode") != "ContentOnly":
            self._request_body["site"].update({"userQuota": str(self._kwargs.get("user_quota"))})
        elif self._kwargs.get("user_quota"):
            self._invalid_parameter_exception()

        self._request_body["site"].update(
            self._get_parameters_dict(self.optional_param_keys, self.optional_param_values)
        )
        return self._request_body

    def get_request(self) -> Dict[str, Any]:
        return self.modified_create_site_request()



from tableau_api_lib import api_endpoints, api_requests, decorators, exceptions, utils
from tableau_api_lib.sample import sample_config
from tableau_api_lib.tableau_server_connection import TableauServerConnection

name = "tableau_api_lib"



"""Defines helper functions for querying details about REST API workbooks and views."""


from io import StringIO
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from tableau_api_lib import TableauServerConnection
from tableau_api_lib.exceptions.tableau_server_exceptions import ContentNotFound
from tableau_api_lib.utils import extract_pages, flatten_dict_column


def get_all_workbook_fields(conn: TableauServerConnection, all_fields: Optional[bool] = True) -> List[Dict[str, Any]]:
    """Returns a list of JSON / dicts describing all available workbooks."""
    fields_param = "_all_" if all_fields is True else "_default_"
    all_workbooks = extract_pages(conn.query_workbooks_for_site, parameter_dict={"fields": f"fields={fields_param}"})
    return all_workbooks


def get_workbooks_dataframe(conn: TableauServerConnection, all_fields: Optional[bool] = False) -> pd.DataFrame:
    """Returns a DataFrame describing all available workbooks. If none are available, an empty DataFrame is returned."""
    try:
        workbooks_df = pd.DataFrame(get_all_workbook_fields(conn=conn, all_fields=all_fields))
    except ContentNotFound:
        workbooks_df = pd.DataFrame()
    return workbooks_df


def get_all_view_fields(
    conn: TableauServerConnection, site_id: str, all_fields: Optional[bool] = True
) -> List[Dict[str, Any]]:
    """Returns a list of JSON / dicts describing all available views."""
    fields_param = "_all_" if all_fields is True else "_default_"
    all_views = extract_pages(
        conn.query_views_for_site,
        content_id=site_id,
        parameter_dict={"fields": f"fields={fields_param}"},
        # parameter_dict={"fields": f"fields={fields_param}", "usage_stats": f"includeUsageStatistics=True"},
    )
    return all_views


def get_views_dataframe(
    conn: TableauServerConnection, site_id: Optional[str] = None, all_fields: Optional[bool] = True
) -> pd.DataFrame:
    """Returns a DataFrame describing all available views. If none are available, an empty DataFrame is returned."""
    if not site_id:
        site_id = conn.site_id
    views_df = pd.DataFrame(get_all_view_fields(conn=conn, site_id=site_id, all_fields=all_fields))
    if not views_df.empty:
        views_df = flatten_dict_column(views_df, keys=["totalViewCount"], col_name="usage")
    return views_df


def get_view_data_dataframe(
    conn: TableauServerConnection, view_id: str, parameter_dict: Optional[Dict[str, Any]] = None, sep: str = ","
) -> pd.DataFrame:
    """Returns a DataFrame containing the data downloaded from a Tableau view."""
    view_data = conn.query_view_data(view_id=view_id, parameter_dict=parameter_dict)
    view_df = pd.read_csv(StringIO(view_data.text), sep=sep)
    return view_df


def get_views_for_workbook_dataframe(conn: TableauServerConnection, workbook_id: str) -> pd.DataFrame:
    """Returns a DataFrame containing details for the views contained within the specified workbook."""
    views_for_workbook = conn.query_views_for_workbook(workbook_id=workbook_id)
    return pd.DataFrame(views_for_workbook.json()["views"]["view"])


def extract_datasource_details(df: pd.DataFrame, keys: List[str]) -> pd.DataFrame:
    """Returns a DataFrame with datasource details unpacked from their original nested column."""
    for key in keys:
        df["datasource_" + key] = df["datasource"].apply(lambda project: project[key])
    df.drop(columns=["datasource"], inplace=True)
    return df


def get_workbook_connections_dataframe(conn: TableauServerConnection, workbook_id: str) -> pd.DataFrame:
    """Returns a DataFrame describing the connections associated with the specified workbook."""
    try:
        connections_json = conn.query_workbook_connections(workbook_id).json()["connections"]["connection"]
        connections_df = pd.DataFrame(connections_json)
        connections_df = flatten_dict_column(connections_df, keys=["id", "name"], col_name="datasource")
    except KeyError:
        connections_df = pd.DataFrame()
    return connections_df


def get_embedded_datasources_dataframe(
    conn: TableauServerConnection,
    workbooks_df: pd.DataFrame,
    workbook_ids: Optional[Union[List[str], pd.Series]] = None,
    id_col: Optional[str] = "id",
    name_col: Optional[str] = "name",
    new_col_prefix: Optional[str] = "",
) -> pd.DataFrame:
    """
    Creates a Pandas DataFrame of all embedded workbook datasources.
    If a subset of workbook IDs are specified, then only embedded connections for those workbooks are returned.

    Args:
        conn: the Tableau Server connection
        workbooks_df: the workbook DataFrame containing details for all workbooks
        workbook_ids: a list of workbook IDs whose embedded datasources will be queried
        id_col: the name of the column containing the workbook ID; defaults to 'id'
        name_col: the name of the column containing the workbook name; defaults to 'name'
        new_col_prefix: the prefix that will be present in all new column names
    """
    workbook_ids = workbook_ids.to_list() if isinstance(workbook_ids, pd.Series) else workbook_ids
    workbook_ids = workbook_ids or []
    if any(workbook_ids):
        workbooks_df = workbooks_df[workbooks_df[id_col].isin(workbook_ids)]
    embedded_datasources_df = pd.DataFrame()
    for index, workbook in workbooks_df.iterrows():
        workbook_connections_df = get_workbook_connections_dataframe(conn=conn, workbook_id=workbook[id_col])
        workbook_connections_df[new_col_prefix + "workbook_name"] = workbook[name_col]
        workbook_connections_df[new_col_prefix + "workbook_id"] = workbook[id_col]
        workbook_connections_df[new_col_prefix + "site_name"] = conn.site_name
        embedded_datasources_df = pd.concat(
            [embedded_datasources_df, workbook_connections_df], ignore_index=True, sort=True
        )
    return embedded_datasources_df



from enum import Enum


class SiteConfig(Enum):
    """Defines constants for site requests."""
    OPTIONAL_PARAM_KEYS = [
        "tierCreatorCapacity",
        "tierExplorerCapacity",
        "tierViewerCapacity",
        "storageQuota",
        "disableSubscriptions",
        "editingFlowsEnabled",
        "schedulingFlowsEnabled",
        "flowsEnabled",
        "guestAccessEnabled",
        "allowSubscriptionAttachments",
        "cacheWarmupEnabled",
        "commentingEnabled",
        "revisionHistoryEnabled",
        "revisionLimit",
        "subscribeOthersEnabled",
        "extractEncryptionMode",
        "requestAccessEnabled",
        "runNowEnabled",
        "dataAlertsEnabled",
        "commentingMentionsEnabled",
        "catalogObfuscationEnabled",
        "flowAutoSaveEnabled",
        "webExtractionEnabled",
        "metricsContentTypeEnabled",
        "notifySiteAdminsOnThrottle",
        "authoringEnabled",
        "customSubscriptionEmailEnabled",
        "customSubscriptionEmail",
        "customSubscriptionFooterEnabled",
        "customSubscriptionFooter",
        "askDataMode",
        "namedSharingEnabled",
        "mobileBiometricsEnabled",
        "sheetImageEnabled",
        "catalogingEnabled",
        "derivedPermissionsEnabled",
        "userVisibilityMode",
        "useDefaultTimeZone",
        "timeZone",
        "autoSuspendRefreshEnabled",
        "autoSuspendRefreshInactivityWindow",
    ]
    OPTIONAL_PARAM_VALUES = [
        "tier_creator_capacity",
        "tier_explorer_capacity",
        "tier_viewer_capacity",
        "storage_quota",
        "disable_subscriptions_flag",
        "editing_flows_enabled_flag",
        "scheduling_flows_enabled_flag",
        "flows_enabled_flag",
        "guest_access_enabled_flag",
        "allow_subscription_attachments_flag",
        "cache_warmup_enabled_flag",
        "commenting_enabled_flag",
        "revision_history_enabled_flag",
        "revision_limit",
        "subscribe_others_enabled_flag",
        "extract_encryption_mode",
        "request_access_enabled_flag",
        "run_now_enabled_flag",
        "data_alerts_enabled_flag",
        "commenting_mentions_enabled_flag",
        "catalog_obfuscation_enabled_flag",
        "flow_auto_save_enabled_flag",
        "web_extraction_enabled_flag",
        "metrics_content_type_enabled_flag",
        "notify_site_admins_on_throttle_flag",
        "authoring_enabled_flag",
        "custom_subscription_email_enabled_flag",
        "custom_subscription_email",
        "custom_subscription_footer_enabled_flag",
        "custom_subscription_footer",
        "ask_data_mode",
        "named_sharing_enabled_flag",
        "mobile_biometrics_enabled_flag",
        "sheet_image_enabled_flag",
        "cataloging_enabled_flag",
        "derived_permissions_enabled_flag",
        "user_visibility_mode",
        "use_default_time_zone_flag",
        "time_zone",
        "auto_suspend_refresh_enabled_flag",
        "auto_suspend_refresh_inactivity_window",
    ]



from types import MethodType
from typing import Any, Dict, List, Optional, Tuple, Union

from typeguard import typechecked

from tableau_api_lib.exceptions import ContentNotFound, PaginationError


def get_page_attributes(query: dict, query_func: MethodType) -> Tuple:
    """Returns page attributes (pageNumber, pageSize, totalAvailable) from a REST API paginated response.

    Args:
        query: The results of the GET request query, containing paginated data.
        query_func: A TableauServerConnection method that will issue a GET request to Tableau Server.

    Returns:
        A tuple describing the active page number, the page size, and the total items available.

    Raises:
        PaginationError: An error triggered when pagination is attempted on a non-paginated object.
    """
    try:
        pagination = query["pagination"]
        page_number = int(pagination["pageNumber"])
        page_size = int(pagination["pageSize"])
        total_available = int(pagination["totalAvailable"])
        return page_number, page_size, total_available
    except KeyError:
        raise PaginationError(query_func)


def extract_pages(
    query_func: object,
    content_id: Optional[str] = None,
    *,
    starting_page: Optional[int] = 1,
    page_size: Optional[int] = 100,
    limit: Optional[int] = None,
    parameter_dict: Optional[Dict[str, Any]] = None,
) -> Union[List[Dict[str, Any]], Dict]:
    """Extracts all available pages from a paginated Tableau Server API response.

    Args:
        query_func: A function that will issue a GET request via the Tableau REST API.
        content_id: The luid for the desired content [group_id, site_id, etc].
        starting_page: The page number to start on. Defaults to the first page (page_number = 1).
        page_size: The maximum number of objects (results) to be returned in any given page.
        limit: The maximum number of objects to return. By default there is no limit.
        parameter_dict: A dict whose values are appended to the REST API URL endpoint as URL parameters.

    Returns:
        A list of JSON / dicts containing the contents of the paginated items.
    """
    parameter_dict = parameter_dict or {}
    extracted_pages = []
    page_number = starting_page
    extracting = True

    while extracting:
        parameter_dict.update({"pageNumber": f"pageNumber={page_number}", "pageSize": f"pageSize={page_size}"})
        query_results = process_query(query_func=query_func, content_id=content_id, parameter_dict=parameter_dict)
        page_number, page_size, total_available = get_page_attributes(query=query_results, query_func=query_func)
        if total_available == 0:
            return [{}]
        extracted_pages, extracting, page_number = update_pagination_params(
            query_results, extracted_pages, page_number, page_size, total_available, limit, extracting
        )
    return extracted_pages


@typechecked
def process_query(query_func: MethodType, content_id: Optional[str], parameter_dict: Dict[str, Any]) -> Dict[Any, Any]:
    """Processes a dynamic GET request via the Tableau REST API.

    Some of the tableau-api-lib methods require a content ID while others will throw an error if an unexpected
    content ID value is passed in when the function is invoked. This function handles both scenarios dynamically.

    Args:
        query_func: The tableau-api-lib method that will be invoked.
        content_id: The luid for the content variety being queried.
        parameter_dict: The dict describing optional additional query parameters.

    Returns:
        The JSON / dict response from the Tableau Server whose content is being queried.

    Raises:
        PaginationError: An error triggered when pagination is attempted on a non-paginated object.
    """
    if content_id:
        try:
            query_results = query_func(content_id, parameter_dict=parameter_dict).json()
        except TypeError:
            raise PaginationError(func=query_func)
    else:
        try:
            query_results = query_func(parameter_dict=parameter_dict).json()
        except TypeError:
            raise PaginationError(func=query_func)
    return query_results


def update_pagination_params(
    query_results: Dict[str, Any],
    extracted_pages: Any,
    page_number: int,
    page_size: int,
    total_available: int,
    limit: int,
    extracting: bool,
) -> Tuple:
    """Updates pagination parameters when iterating through all available pages for a variety of content.

    Args:
        query_results: The JSON / dict response received from Tableau Server following a GET request.
        extracted_pages: The pages that have been extracted from the REST API request.
        page_number: tbd
        page_size: tbd
        total_available: tbd
        limit: tbd
        extracting: tbd

    Returns:
        Pagination parameters describing active state of the pagination process.

    Raises:
        ContentNotFound: An exception thrown when no content of the variety queried exists on the Tableau Server.
    """
    try:
        outer_key = [key for key in query_results.keys() if key != "pagination"].pop()
        inner_key = list(query_results[outer_key].keys()).pop()
        extracted_pages += query_results[outer_key][inner_key]
    except IndexError:
        raise ContentNotFound()

    if limit:
        if limit <= len(extracted_pages):
            extracted_pages = extracted_pages[:limit]
            extracting = False
    elif total_available <= (page_number * page_size):
        extracting = False
    else:
        page_number += 1
    return extracted_pages, extracting, page_number



from typing import Optional, Any, Dict

from tableau_api_lib.api_endpoints import BaseEndpoint


class PermissionsEndpoint(BaseEndpoint):
    def __init__(
        self,
        ts_connection,
        add_object_permissions: bool = False,
        query_object_permissions: bool = False,
        delete_object_permissions: bool = False,
        object_type: Optional[str] = None,
        object_id: Optional[str] = None,
        query_default_project_permissions: bool = False,
        add_default_project_permissions: bool = False,
        delete_default_project_permissions: bool = False,
        project_permissions_object: Optional[str] = None,
        delete_permissions_object: Optional[str] = None,
        delete_permissions_object_id: Optional[str] = None,
        capability_name: Optional[str] = None,
        capability_mode: Optional[str] = None,
        project_id: Optional[str] = None,
        parameter_dict: Optional[Dict[str, Any]] = None,
    ):
        """Builds API endpoints for REST API permissions methods.

        Args:
            ts_connection: The Tableau Server connection object.
            add_object_permissions: True if adding object permissions, False otherwise.
            query_object_permissions: True if querying object permissions, False otherwise.
            delete_object_permissions: True if deleting object permissions, False otherwise.
            object_type: the Tableau object type (workbook, etc).
            object_id: the Tableau object ID.
            query_default_project_permissions: True if querying default project permissions, False otherwise.
            add_default_project_permissions: True if adding default project permissions, False otherwise.
            delete_default_project_permissions: True if deleting default project permissions, False otherwise.
            project_permissions_object: the project permissions object (workbook, etc).
            delete_permissions_object: the permissions object to delete (user, group, etc).
            delete_permissions_object_id: the ID of the permissions object to delete (user ID, group ID, etc).
            capability_name: the permissions capability being targeted (only use when deleting).
            capability_mode: set this value to 'Allow' when you want to delete a permission allowing a
                capability; set this value to 'Deny' when you want to delete a permission denying a capability.
            project_id: the project ID.
            parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
                literal text that will be appended to the URL endpoint.
        """

        super().__init__(ts_connection)
        self._add_object_permissions = add_object_permissions
        self._query_object_permissions = query_object_permissions
        self._delete_object_permissions = delete_object_permissions
        self._object_type = self._enforce_plurals(object_type, self.valid_project_permissions_objects)
        self._object_id = object_id
        self._query_default_project_permissions = query_default_project_permissions
        self._add_default_project_permissions = add_default_project_permissions
        self._delete_default_project_permissions = delete_default_project_permissions
        self._project_permissions_object = self._enforce_plurals(
            project_permissions_object, self.valid_project_permissions_objects
        )
        self._delete_permissions_object = self._enforce_plurals(
            delete_permissions_object, self.valid_project_delete_permissions_objects
        )
        self._delete_permissions_object_id = delete_permissions_object_id
        self._capability_name = capability_name
        self._capability_mode = capability_mode
        self._project_id = project_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._add_object_permissions,
            self._query_object_permissions,
            self._delete_object_permissions,
            self._query_default_project_permissions,
            self._add_default_project_permissions,
            self._delete_default_project_permissions,
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def valid_project_permissions_objects(self):
        return [
            "datasources",
            "datasource",
            "workbooks",
            "workbook",
            "flows",
            "flow",
            "views",
            "view",
            "projects",
            "project",
            "metrics",
            "metric",
            "datarole",
            "dataroles",
            "lens",
            "lenses"
        ]

    @property
    def valid_project_delete_permissions_objects(self):
        return ["users", "user", "groups", "group"]

    def _enforce_plurals(self, permissions_object, valid_permissions_objects):
        if permissions_object:
            if permissions_object in valid_permissions_objects:
                if permissions_object[-1] == "s":
                    return permissions_object
                else:
                    return permissions_object + "s"
            else:
                self._invalid_parameter_exception()

    @property
    def base_permissions_url(self):
        return "{0}/api/{1}/sites/{2}".format(
            self._connection.server, self._connection.api_version, self._connection.site_id
        )

    @property
    def base_object_permissions_url(self):
        return "{0}/{1}/{2}/permissions".format(self.base_permissions_url, self._object_type, self._object_id)

    @property
    def base_query_default_permissions_url(self):
        return "{0}/projects/{1}/default-permissions/{2}".format(
            self.base_permissions_url, self._project_id, self._project_permissions_object
        )

    @property
    def base_delete_permission_url(self):
        return "{0}/{1}/{2}/{3}/{4}".format(
            self.base_object_permissions_url,
            self._delete_permissions_object,
            self._delete_permissions_object_id,
            self._capability_name,
            self._capability_mode,
        )

    @property
    def base_delete_default_permissions_url(self):
        return "{0}/{1}/{2}/{3}/{4}".format(
            self.base_query_default_permissions_url,
            self._delete_permissions_object,
            self._delete_permissions_object_id,
            self._capability_name,
            self._capability_mode,
        )

    def get_endpoint(self):
        url = None
        if self._add_object_permissions:
            url = self.base_object_permissions_url
        elif self._query_object_permissions:
            url = self.base_object_permissions_url
        elif self._delete_object_permissions:
            url = self.base_delete_permission_url
        elif self._delete_default_project_permissions:
            url = self.base_delete_default_permissions_url
        elif self._query_default_project_permissions:
            url = self.base_query_default_permissions_url
        elif self._add_default_project_permissions:
            url = self.base_query_default_permissions_url
        else:
            self._invalid_parameter_exception()

        return self._append_url_parameters(url)



from typing import Any, Dict, List

from tableau_api_lib.exceptions import InvalidParameterException
from tableau_api_lib.tableau_server_connection import TableauServerConnection


class BaseRequest:
    """The base request from which all other API requests are built."""

    def __init__(self, ts_connection: TableauServerConnection):

        self._connection = ts_connection
        self._request_body = {}

    @staticmethod
    def _get_parameters_dict(param_keys: List[str], param_values: List[Any]) -> Dict[str, Any]:
        """Returns zipped keys and values into a dict, but only for pairs where the param value is truthy."""
        params_dict = {}
        for i, key in enumerate(param_keys):
            if param_values[i] or (param_values[i] is False):
                params_dict.update({key: param_values[i]})
        return params_dict

    def _invalid_parameter_exception(self) -> None:
        raise InvalidParameterException(class_name=self.__class__.__name__, parameters=vars(self))



"""
workflow logic:
1) get workbook details for the source site
2) get workbook details for the target site
3) get overlapping workbooks (source & target)
4) if overwrite_policy == 'overwrite', then delete the overlapping workbooks on the target site
5) if overwrite_policy == None, then throw an exception
6) if overwrite_policy == 'merge', then proceed with cloning the non-overlapping workbooks and leave the others as is
7) for each workbook on the source site, build a clone of that workbook on the target site
    7a) for each workbook, query its connections
    7b) for each connection, check if it is a published data source
    7c) build a dataframe that labels datasources as embedded or published
    7d) for each published data source on the source site, download and publish that data source to the target site
            must open each file and edit its server, username, and site name
    7e) make sure each data source is published to the correct respective workbook
    7f) download all workbooks (will be downloaded as .twbx)
        save each workbook to the local file system (in the future provide a way to save to cloud storage)
    7g) publish all 'published' datasources to the target site
    7h) publish all workbooks to their respective workbooks, with published data source credentials
            for published data sources: server name, connection username, connection password must be provided

"""


import pandas as pd
import os

from tableau_api_lib.utils.querying import get_workbooks_dataframe, get_projects_dataframe, \
    get_embedded_datasources_dataframe
from tableau_api_lib.utils.filemod import modify_tableau_zipfile, set_temp_dirs, delete_temp_files, \
    replace_unzipped_xml_file
from tableau_api_lib.utils import flatten_dict_column, get_server_netloc
from tableau_api_lib.exceptions import ContentOverwriteDisabled


def get_source_workbook_df(conn_source, workbook_names=None):
    """
    Query details for all workbooks on the source site, or only the workbooks listed in 'workbook_names'.
    :param class conn_source: the Tableau Server connection object
    :param list workbook_names: a list of the desired workbook names whose details will be queried
    :return: Pandas DataFrame
    """
    workbook_df = get_workbooks_dataframe(conn_source)
    workbook_df.fillna(value='', inplace=True)
    if workbook_names:
        workbook_df = workbook_df[workbook_df['name'].isin(workbook_names)]
    workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name'], col_name='project')
    workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name', 'email'], col_name='owner')
    workbook_df.columns = ["source_" + column for column in workbook_df.columns]
    workbook_df['source_project_name_lower'] = workbook_df['source_project_name'].apply(lambda x: str(x).lower())
    return workbook_df


def get_target_workbook_df(conn_target, workbook_names=None):
    """
    Query details for all workbooks on the source site, or only the workbooks listed in 'workbook_names'.
    :param class conn_target: the Tableau Server connection object
    :param list workbook_names: a list of the desired workbook names whose details will be queried
    :return: Pandas DataFrame
    """
    try:
        workbook_df = get_workbooks_dataframe(conn_target)
        workbook_df.fillna(value='', inplace=True)
        if workbook_names:
            workbook_df = workbook_df[workbook_df['name'].isin(workbook_names)]
        workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name'], col_name='project')
        workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name', 'email'], col_name='owner')
        workbook_df.columns = ["target_" + column for column in workbook_df.columns]
        workbook_df['target_project_name_lower'] = workbook_df['target_project_name'].apply(lambda x: str(x).lower())
    except KeyError:
        workbook_df = pd.DataFrame()
    return workbook_df


def get_source_to_target_wb_df(source_workbook_df, conn_target):
    """
    Get a DataFrame with source workbook details, source user details, and source target details combined.
    :param pd.DataFrame source_workbook_df: the source workbook details
    :param class conn_target: the target Tableau Server connection
    :return DataFrame: workbook_details_df
    """
    target_projects_df = get_projects_dataframe(conn_target)
    target_projects_df.columns = ["target_project_" + column for column in target_projects_df.columns]
    target_projects_df['target_project_name_lower'] = target_projects_df['target_project_name'].apply(lambda x: str(x).lower())
    workbook_details_df = source_workbook_df.merge(target_projects_df,
                                                   how='left',
                                                   left_on='source_project_name_lower',
                                                   right_on='target_project_name_lower')
    return workbook_details_df


def get_workbook_connections_df(conn_source, conn_target, workbooks_df):
    workbook_connections_df = get_embedded_datasources_dataframe(conn_source,
                                                                 workbooks_df=workbooks_df,
                                                                 id_col='source_id',
                                                                 name_col='source_name',
                                                                 new_col_prefix='source_')
    workbook_connections_df['serverAddress'] = workbook_connections_df['serverAddress'].\
        apply(lambda x: x.replace(get_server_netloc(conn_source.server), get_server_netloc(conn_target.server)))
    workbook_connections_df['serverPort'] = workbook_connections_df['serverPort'].fillna(0).astype('int').astype('str')
    workbook_connections_df['userName'] = workbook_connections_df['userName'].fillna('')
    return workbook_connections_df


def get_workbook_credentials_df(workbook_connections_df, credentials_file_path=None):
    if not credentials_file_path:
        workbook_connections_df['password'] = ''
    else:
        credentials_df = pd.read_csv(credentials_file_path)
        credentials_df['password'].fillna('', inplace=True)
        credentials_df['serverPort'].fillna(0, inplace=True)
        # credentials_df['password'].fillna('', inplace=True)
        credentials_df['serverPort'] = credentials_df['serverPort'].astype('int').astype('str')
        workbook_connections_df = workbook_connections_df.merge(credentials_df,
                                                                how='left',
                                                                on=['serverAddress', 'userName'],
                                                                suffixes=(None, '_delete'))
    workbook_connections_df['password'] = workbook_connections_df['password'].fillna('')
    credential_cols = [col for col in list(workbook_connections_df.columns) if not str(col).endswith('_delete')]
    return workbook_connections_df[credential_cols]


def get_project_workbook_credentials_df(workbook_credentials_df, source_to_target_wb_df):
    source_to_target_abbreviated_df = source_to_target_wb_df[['source_id', 'source_project_name', 'target_project_name',
                                                              'source_project_name_lower', 'target_project_name_lower']]
    project_workbook_credentials_df = workbook_credentials_df.merge(source_to_target_abbreviated_df,
                                                                    how='left',
                                                                    left_on='source_workbook_id',
                                                                    right_on='source_id')
    return project_workbook_credentials_df


def get_workbook_names(workbook_df, column_name) -> list:
    """
    Get the workbook names for the active site on the specified Tableau Server connection.
    :param DataFrame workbook_df: the workbook details DataFrame for the server
    :param str column_name: the name of the column containing workbook names [source_workbook_name, target_workbook_name]
    :return: list
    """
    workbook_names = list(workbook_df[column_name])
    return workbook_names


def get_overlapping_workbook_names(source_workbook_df, target_workbook_df) -> list:
    try:
        source_workbook_names = set(get_workbook_names(source_workbook_df, 'source_workbook_name'))
        target_workbook_names = set(get_workbook_names(target_workbook_df, 'target_workbook_name'))
        overlapping_workbook_names = source_workbook_names.intersection(target_workbook_names)
        return list(overlapping_workbook_names)
    except KeyError:
        return []


def get_workbook_file_type(response):
    try:
        file_type = ''
        content_type = dict(response.headers)['Content-Type']
        if str(content_type.lower()) == 'application/xml':
            file_type = 'twb'
        elif str(content_type.lower()) == 'application/octet-stream':
            file_type = 'twbx'
        return file_type
    except KeyError:
        raise Exception(f"An error occurred while downloading the workbook: status code {response.status_code}")


def download_workbooks(conn_source, workbooks_df, download_dir=None, target_project_dir=None):
    download_dir = download_dir or os.getcwd() + '/temp/workbooks/source'
    target_project_dir = target_project_dir or os.getcwd() + '/temp/workbooks/target'
    workbook_file_names_df = pd.DataFrame()
    for index, workbook in workbooks_df.iterrows():
        response = conn_source.download_workbook(workbook_id=workbook['source_id'])
        file_type = get_workbook_file_type(response)
        print("downloading workbook '{0}' to '{1}/{0}.{2}'...".format(workbook['source_name'],
                                                                      download_dir,
                                                                      file_type))
        with open(f"{download_dir}/{workbook['source_name']}.{file_type}", 'wb') as file:
            file.write(response.content)
        workbook_file_name_df = workbook[['source_name', 'source_project_name']].copy()
        workbook_file_name_df['file_path'] = f"{target_project_dir}/{workbook['source_name']}.{file_type}"
        workbook_file_names_df = pd.concat([workbook_file_names_df, workbook_file_name_df], ignore_index=True, sort=False)
    return workbook_file_names_df


def add_downloaded_file_names(workbooks_df, downloaded_file_names_df):
    workbooks_df = workbooks_df.merge(downloaded_file_names_df,
                                      how='left',
                                      on=['source_name', 'source_project_name'])
    return workbooks_df


def modify_workbooks_by_project(conn_source, conn_target, source_project_dir, target_project_dir, extraction_dir):
    for file in os.listdir(source_project_dir):
        file_type = os.path.splitext(file)[-1]
        if file_type == '.twbx':
            modify_tableau_zipfile(zipfile_path=source_project_dir + '/' + file,
                                   conn_source=conn_source, conn_target=conn_target,
                                   extraction_dir_path=extraction_dir,
                                   destination_dir_path=target_project_dir)
        elif file_type == '.twb':
            replace_unzipped_xml_file(file_path=source_project_dir + '/' + file,
                                      conn_source=conn_source, conn_target=conn_target,
                                      extraction_dir_path=target_project_dir)
        delete_temp_files(extraction_dir)


def delete_workbooks(conn, workbook_details_df, workbook_names):
    print("deleting overlapping target workbooks...")
    responses = []
    workbooks_to_delete = workbook_details_df[workbook_details_df['target_workbook_name'].isin(workbook_names)]
    for index, workbook in workbooks_to_delete.iterrows():
        print("deleting workbook: (name='{}', id='{}'".
              format(workbook['target_workbook_id'], workbook['target_workbook_name']))
        responses.append(conn.delete_workbook(workbook_id=workbook['target_workbook_id']))
    print("overlapping target workbooks deleted")
    return conn


def validate_inputs(overwrite_policy):
    valid_overwrite_policies = [
        None,
        'overwrite',
        'merge'  # add logic for merging (leave existing content as is and create any missing content)
    ]
    if overwrite_policy not in valid_overwrite_policies:
        raise ValueError("Invalid overwrite policy provided: '{}'".format(overwrite_policy))


def process_overlapping_workbooks(conn_target,
                                  target_workbook_df,
                                  overlapping_workbook_names,
                                  overwrite_policy) -> None:
    """
    Processes overlapping workbooks depending on the overwrite policy specified.
    :param class conn_target: the target Tableau Server connection
    :param pd.DataFrame target_workbook_df: a Pandas DataFrame with target workbook details
    :param list overlapping_workbook_names: a list of workbooks present on both the source and target connections
    :param str overwrite_policy: must be set to 'overwrite' to enable overwriting existing content
    :return:
    """
    if any(overlapping_workbook_names) and not overwrite_policy:
        raise ContentOverwriteDisabled('workbooks')
    if any(overlapping_workbook_names) and overwrite_policy == 'overwrite':
        delete_workbooks(conn_target, target_workbook_df, overlapping_workbook_names)
    pass


def publish_workbooks_by_project(conn_target, project_workbooks_df, project_workbook_credentials_df):
    for index, workbook in project_workbooks_df.iterrows():
        workbook_credentials_df = project_workbook_credentials_df.loc[
            project_workbook_credentials_df['source_workbook_name'] == workbook['source_name']]
        response = conn_target.publish_workbook(workbook_file_path=workbook['file_path'],
                                                workbook_name=workbook['source_name'],
                                                project_id=workbook['target_project_id'],
                                                server_address=workbook_credentials_df['serverAddress'].to_list(),
                                                port_number=workbook_credentials_df['serverPort'].to_list(),
                                                connection_username=workbook_credentials_df['userName'].to_list(),
                                                connection_password=workbook_credentials_df['password'].to_list(),
                                                embed_credentials_flag=workbook_credentials_df['embedPassword'].to_list())
        print("publish response for {}:\n{}".format(workbook['source_name'], response.json()))
        if response.status_code in [200, 201]:
            project_workbooks_df.at[index, 'target_id'] = response.json()['workbook']['id']
    return project_workbooks_df


def clone_workbooks_by_project(conn_source,
                               conn_target,
                               source_to_target_wb_df,
                               project_workbook_credentials_df,
                               temp_dir,
                               project,
                               extraction_dir):
    source_project_dir = f"{temp_dir}/workbooks/source/{project}"
    target_project_dir = f"{temp_dir}/workbooks/target/{project}"
    os.makedirs(source_project_dir, exist_ok=False)
    os.makedirs(target_project_dir, exist_ok=False)
    project_workbooks_df = source_to_target_wb_df[
        source_to_target_wb_df['target_project_name_lower'] == str(project).lower()].copy()
    project_workbook_credentials_df = project_workbook_credentials_df[
        project_workbook_credentials_df['target_project_name_lower'] == str(project).lower()].copy()
    workbook_file_names_df = download_workbooks(conn_source, project_workbooks_df, download_dir=source_project_dir, target_project_dir=target_project_dir)
    project_workbooks_df = add_downloaded_file_names(project_workbooks_df, workbook_file_names_df)
    modify_workbooks_by_project(conn_source, conn_target, source_project_dir, target_project_dir, extraction_dir)
    project_workbooks_df.to_csv(project + '_project_workbooks_df.csv')
    project_workbook_credentials_df.to_csv(project + '_project_workbook_credentials_df.csv')
    print("project_workbooks_df:\n", project_workbooks_df['file_path'])
    print("project_workbook_credentials_df:\n", project_workbook_credentials_df)
    published_workbooks = publish_workbooks_by_project(conn_target, project_workbooks_df, project_workbook_credentials_df)
    print("published_workbooks_df:\n", published_workbooks)
    # project_datasources = publish_datasources_by_project(conn_target, project_datasources_df)
    # update_datasources_by_project(conn_target, project_datasources)


# TODO hide sheets not visible on the source server when publishing to the target; copy tags and owner (make optional)
def clone_workbooks(conn_source,
                    conn_target,
                    workbook_names=None,
                    credentials_file_path=None,
                    temp_dir=None,
                    overwrite_policy=None):
    """
    Clones workbooks from the source server to the target server.
    :param class conn_source: the source server connection
    :param class conn_target: the target server connection
    :param list workbook_names: (optional) a list of workbook names to clone; specifying no names clones all workbooks
    :param str temp_dir: (optional) designate the location where temp files will be stored
    :param str overwrite_policy: (optional) set to 'overwrite' to overwrite content; defaults to not overwriting
    :return: None
    """
    temp_dir, extraction_dir = set_temp_dirs(temp_dir)
    validate_inputs(overwrite_policy)
    source_workbook_df = get_source_workbook_df(conn_source=conn_source, workbook_names=workbook_names)
    target_workbook_df = get_target_workbook_df(conn_target=conn_target, workbook_names=workbook_names)
    # refactor the overlapping checks to a new function: process_overwrite_policy()
    if any(target_workbook_df.columns):
        overlapping_workbook_names = get_overlapping_workbook_names(source_workbook_df, target_workbook_df)
        process_overlapping_workbooks(conn_target, target_workbook_df, overlapping_workbook_names, overwrite_policy)
    source_to_target_wb_df = get_source_to_target_wb_df(source_workbook_df=source_workbook_df, conn_target=conn_target)
    workbook_conn_df = get_workbook_connections_df(conn_source, conn_target, source_to_target_wb_df)
    workbook_credentials_df = get_workbook_credentials_df(workbook_conn_df, credentials_file_path)
    project_workbook_credentials_df = get_project_workbook_credentials_df(workbook_credentials_df, source_to_target_wb_df)
    try:
        for project in source_to_target_wb_df['source_project_name'].unique():
            clone_workbooks_by_project(conn_source=conn_source,
                                       conn_target=conn_target,
                                       source_to_target_wb_df=source_to_target_wb_df,
                                       project_workbook_credentials_df=project_workbook_credentials_df,
                                       temp_dir=temp_dir,
                                       project=project,
                                       extraction_dir=extraction_dir)
    finally:
        delete_temp_files(temp_dir)
    # -> update workbooks to have the correct owner (and other metadata)
    #   -> optional mapping stage: map_workbook_owners (searches for original owner on target site, otherwise default)



from functools import wraps

from tableau_api_lib.exceptions import InvalidParameterException


def validate_schedule_state_override(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        state_override_value = kwargs['state_override_value'] if 'state_override_value' in kwargs.keys() else None
        state_override_value = state_override_value.lower().capitalize() if state_override_value else None
        if state_override_value in ['Suspended', 'Active', None]:
            return func(*args, **kwargs)
        else:
            raise InvalidParameterException(func.__name__, state_override_value)
    return wrapper



"""Helper functions for querying REST API data for extract refresh tasks."""


from typing import Any, Dict, List

import pandas as pd

from tableau_api_lib import TableauServerConnection
from tableau_api_lib.exceptions.tableau_server_exceptions import ContentNotFound


def get_extract_refresh_tasks_for_site(conn: TableauServerConnection) -> List[Dict[str, Any]]:
    """Returns a list of Python dicts describing all extract refresh tasks for the active site."""
    try:
        return conn.get_extract_refresh_tasks_for_site().json()["tasks"]["task"]
    except KeyError:
        raise ContentNotFound(content_type="extract refresh tasks")


def get_extract_refresh_tasks_dataframe(conn: TableauServerConnection) -> pd.DataFrame:
    """Returns a DataFrame describing all extract refresh tasks for the site."""
    try:
        extract_refresh_tasks = [task["extractRefresh"] for task in get_extract_refresh_tasks_for_site(conn)]
        extract_refresh_tasks_df = pd.DataFrame(extract_refresh_tasks)
    except ContentNotFound:
        extract_refresh_tasks_df = pd.DataFrame()
    return extract_refresh_tasks_df



import pandas as pd


def get_webhooks_dataframe(conn) -> pd.DataFrame:
    """
    Returns a Pandas DataFrame containing details for all webhooks on the active site.
    :param class conn: the Tableau Server connection
    :return: pd.DataFrame
    """
    try:
        webhooks_json = conn.query_webhooks().json()['webhooks']['webhook']
    except KeyError:
        webhooks_json = {}
    webhooks_df = pd.DataFrame(webhooks_json)
    return webhooks_df



"""
The functions below enable you to clone the schedules from Server A to Server B.
This is particularly useful when doing site exports / imports.
The only functions are for use outside of this module:

clone_schedules
clone_schedule_state
override_schedule_state
"""
import pandas as pd
import warnings

from tableau_api_lib.utils.querying import get_schedules_dataframe
from tableau_api_lib.decorators import validate_schedule_state_override
from tableau_api_lib.exceptions import InvalidParameterException


def get_schedule_details(conn, schedule_names=None):
    schedules_df = get_schedules_dataframe(conn)
    if schedule_names:
        schedules_df = schedules_df.loc[schedules_df['name'].isin(schedule_names)]
    schedule_ids = list(schedules_df['id'])
    schedule_details = [conn.update_schedule(schedule_id).json()['schedule'] for schedule_id in schedule_ids]
    return pd.DataFrame(schedule_details)


def create_base_schedules(conn,
                          schedules_df,
                          prefix=None,
                          suffix=None):
    prefix = prefix or ''
    suffix = suffix or ''
    responses = []
    for index, schedule in schedules_df.iterrows():
        response = conn.create_schedule(schedule_name=prefix + schedule['name'] + suffix,
                                        schedule_priority=schedule['priority'],
                                        schedule_type=schedule['type'],
                                        schedule_execution_order=schedule['executionOrder'],
                                        schedule_frequency=schedule['frequency'],
                                        start_time=schedule['frequencyDetails']['start'],
                                        end_time=schedule['frequencyDetails'].get('end', None),
                                        interval_expression_list=schedule['frequencyDetails']['intervals']['interval']
                                        if schedule['frequencyDetails'].get('intervals') else None)
        try:
            responses.append(response.json()['schedule'])
        except KeyError:
            warnings.warn(f"""
            There is already a schedule named '{prefix + schedule['name'] + suffix}'. 
            Duplicates .
            """)
    return responses


def update_schedule_state(conn,
                          source_schedules_df,
                          destination_schedules_df):
    responses = []
    for index, schedule in destination_schedules_df.iterrows():
        response = conn.update_schedule(schedule_id=schedule['id'],
                                        schedule_state=source_schedules_df.iloc[index]['state'])
        try:
            responses.append(response.json()['schedule'])
        except KeyError:
            warnings.warn(f"Unable to update schedule '{schedule['name']}'")
    return pd.DataFrame(responses)


@validate_schedule_state_override
def override_schedule_state(conn, state_override_value=None, schedule_names=None):
    """
    Replace the existing schedule state (Active or Suspended) with the specified state override value.
    :param TableauServerConnection conn: the Tableau Server connection
    :param str state_override_value: specify a state override value ('Active' or 'Suspended')
    :param list schedule_names: specify schedule names if overriding a subset of existing schedules
    :return: pd.DataFrame
    """
    if state_override_value:
        responses = []
        schedules_df = get_schedule_details(conn, schedule_names)
        for index, schedule in schedules_df.iterrows():
            response = conn.update_schedule(schedule_id=schedule['id'], schedule_state=state_override_value)
            try:
                responses.append(response.json()['schedule'])
            except KeyError:
                warnings.warn(f"Unable to update schedule '{schedule['name']}'")
        return pd.DataFrame(responses)
    else:
        return pd.DataFrame()


def verify_schedule_names_align(source_schedule_names, destination_schedule_names, func_name=None):
    if source_schedule_names and destination_schedule_names:
        if len(source_schedule_names) != len(destination_schedule_names):
            print("When specifying schedule names, \
            you must provide an equal number of names for both the origin and destination.")
            raise InvalidParameterException(func_name,
                                            (source_schedule_names, destination_schedule_names))
    elif source_schedule_names or destination_schedule_names:
        raise InvalidParameterException(func_name,
                                        (source_schedule_names, destination_schedule_names))


def clone_schedule_state(conn_source,
                         conn_destination,
                         source_schedule_names=None,
                         destination_schedule_names=None) -> pd.DataFrame:
    """
    Clone schedule states from Tableau Server environment 'conn_source' to 'conn_destination'.
    :param TableauServerConnection conn_source: the source Tableau Server connection
    :param TableauServerConnection conn_destination: the destination Tableau Server connection
    :param list source_schedule_names: (optional) define a subset of source schedule names
    :param list destination_schedule_names: (optional) define a subset of destination schedule names
    :return: pd.DataFrame
    """
    verify_schedule_names_align(source_schedule_names, destination_schedule_names, func_name='clone_schedule_state()')
    source_schedule_df = get_schedule_details(conn_source, schedule_names=source_schedule_names)
    destination_schedule_df = get_schedule_details(conn_destination, schedule_names=destination_schedule_names)
    updated_schedules = update_schedule_state(conn_destination, source_schedule_df, destination_schedule_df)
    return updated_schedules


@validate_schedule_state_override
def clone_schedules(conn_source,
                    conn_destination,
                    state_override_value=None,
                    schedule_names=None,
                    prefix=None,
                    suffix=None) -> pd.DataFrame:
    """
    Clone schedules from Tableau Server environment 'conn_source' to 'conn_destination'.
    :param TableauServerConnection conn_source: the source Tableau Server connection
    :param TableauServerConnection conn_destination: the destination Tableau Server connection
    :param str state_override_value: (optional) define a state value to override schedule states from the source
    :param list schedule_names: (optional) define a subset of schedule names to clone
    :param str prefix: (optional) define a prefix to add to the cloned schedule names
    :param str suffix: (optional) define a suffix to add to the cloned schedule names
    :return: pd.DataFrame
    """
    source_schedule_df = get_schedule_details(conn=conn_source, schedule_names=schedule_names)
    base_schedules_responses = create_base_schedules(conn_destination, source_schedule_df, prefix, suffix)
    base_schedule_names = [schedule['name'] for schedule in base_schedules_responses]

    destination_schedule_df = get_schedule_details(conn=conn_destination, schedule_names=base_schedule_names)
    cloned_schedules = update_schedule_state(conn_destination, source_schedule_df, destination_schedule_df)
    if state_override_value:
        cloned_schedules = override_schedule_state(conn_destination, state_override_value, base_schedule_names)
    return cloned_schedules
